#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "em_mog.h"
#include "bg_model.h"
#include "k_means.h"
#include "macros.h"

static char* bg_em_cmd[NUM_BG_EM_COMMAND] = {
		"ngaussian",
		"nth mean",
		"nth weight",
		"nth matrix",
		"epsilon",
		"noise level",
};

static char* bg_k_means_cmd[NUM_BG_K_MEANS_COMMAND] = {
		"nmean",
		"nth mean"
};

struct circle_buff *
new_circle_buff()
{
	struct circle_buff *ptr;
	ptr = malloc(sizeof(struct circle_buff));
	return_val_if_fail(ptr != NULL, ptr);
	return ptr;
}

/**
 @function
 set_circle_buff_data_size - set data dimension and buffer size from c_buff
 @param dim    : data dimension
 @param ndata  : circle buffer size
 @param c_buff : pointer to struct circle_buff
 @note: if define DEBUG in this function done check that the c_buff is valid
 */
void
set_circle_buff_data_size(int dim, int ndata, struct circle_buff *c_buff)
{
	return_if_fail(c_buff != NULL);
	c_buff->dim       = dim;
	c_buff->buff_size = ndata;
}

void
init_circle_buff(struct circle_buff *c_buff)
{
	return_if_fail(c_buff != NULL);
	int ndata, dim;
	ndata = c_buff->buff_size;
	dim   = c_buff->dim;
	c_buff->head = c_buff->tail = 0;
	c_buff->buff = malloc(ndata * dim * sizeof(double));
}

void
clear_circle_buff(struct circle_buff *c_buff)
{
	return_if_fail(c_buff != NULL);
	return_if_fail(c_buff->buff != NULL);
	free(c_buff->buff);
	c_buff->buff_size = 0;
	c_buff->dim = 0;
	c_buff->head = 0;
	c_buff->tail = 0;
}

void
write_circle_buff(struct circle_buff *c_buff, const double *data)
{
	return_if_fail(c_buff != NULL);
	return_if_fail(c_buff->buff != NULL);
	return_if_fail(data != NULL);

	int ndata, dim;
	ndata = c_buff->buff_size;
	dim   = c_buff->dim;

	//memcpy(c_buff->buff + c_buff->head * dim, data, dim * sizeof(double));
	double *ptr1;
	const double *ptr2;
	ptr1 = c_buff->buff + c_buff->head * dim;
	ptr2 = data;
	*ptr1++ = *ptr2++;
	*ptr1++ = *ptr2++;
	*ptr1 = *ptr2;
	if (c_buff->tail + 1 != ndata)
		c_buff->tail++;
	c_buff->head++;
	if (c_buff->head == ndata)
		c_buff->head = 0;
	//c_buff->head %= ndata;
}

void
reset_circle_buff(struct circle_buff *c_buff)
{
	return_if_fail(c_buff != NULL);
	return_if_fail(c_buff->buff != NULL);
	c_buff->head = c_buff->tail = 0;
}

void
free_circle_buff(struct circle_buff **c_buff)
{
	return_if_fail(c_buff != NULL);
	return_if_fail(*c_buff != NULL);
	double *tmp = (*c_buff)->buff;
	return_if_fail(tmp != NULL);

	free(tmp);
	free(*c_buff);
	*c_buff = NULL;
}

static void
bg_model_adaptive_init_rgb(double *means, int dim, int k, const double *data, int data_size)
{
	const int n = 256;
	double hist[n], value[n];
	int idxs[n];
	int i, l, r, front, gray;
	double min, tmp, *ptr;
	double const *ptr1;
	memset(hist, 0, sizeof(double) * n);

	ptr1 = data;

	for (i = 0; i < n; i++) {
		idxs[i] = -1;
	}

	for (i = 0; i < data_size; i++) {
		gray = (int)(0.299 * ptr1[0] + 0.587 * ptr1[1] + 0.114 * ptr1[2]);
		return_if_fail(gray < 0 || gray > 255);
		hist[gray]++;
		idxs[gray] = i;
		ptr += dim;
	}

	for (i = 0; i < n; i++) {
		hist[i] /= data_size;
	}

	memcpy(value, hist, n * sizeof(double));

	for (i = 0; i < n - 1; i++) {
		hist[i+1] += hist[i];
	}

	min = n + 1;
	front = 0;
	for (i = 0; i < n - 1; i++) {
		tmp = i * hist[i] + (n - i) * (1 - hist[i + 1]);
		if (tmp < min) {
			min = tmp;
			front = i;
			continue;
		}
		if (tmp == min) {
			if (hist[i] > hist[front]) {
				front = i;
			}
		}
	}
	tmp = -1;
	l = 0;
	for (i = 0; i < front; i++) {
		if (value[i] > tmp) {
			tmp = value[i];
			l = i;
		}
	}

	tmp = -1;
	r = n - 1;
	for (i = front; i < n; i++) {
		if (value[i] > tmp) {
			tmp = value[i];
			r = i;
		}
	}
	if (idxs[l] == -1)
		l = r;
	if (idxs[r] == -1)
		r = l;

	if (idxs[l] == -1 || idxs[r] == -1) {
		FILE *fp = fopen("error.txt", "w");
		fprintf(fp, "%d %d\n", r, l);
		for (i = 0; i < n; i++) {
			fprintf(fp, " % f %f %d\n", hist[i], value[i], idxs[i]);
		}
		fclose(fp);
		abort();
	}
	ptr = means;
	ptr1 = data + idxs[l] * dim;
	for (i = 0; i < dim; i++) {
		*ptr = *ptr1;
		ptr++;
		ptr1++;
	}
	ptr1 = data + idxs[r] * dim;
	for (i = 0; i < dim; i++) {
		*ptr = *ptr1;
		ptr++;
		ptr1++;
	}
}

static void
bg_model_adaptive_init_rgb_2_means(double *means, int dim, int k, const double *data, int data_size)
{
	const int n = 256;
	double hist[n], value[n];
	int idxs[n];
	int i, l, r, front, gray;
	double min, tmp, *ptr;
	double const *ptr1;
	memset(hist, 0, sizeof(double) * n);

	ptr1 = data;

	for (i = 0; i < n; i++) {
		idxs[i] = -1;
	}

	for (i = 0; i < data_size; i++) {
		gray = (int)(0.299 * ptr1[0] + 0.587 * ptr1[1] + 0.114 * ptr1[2]);
		return_if_fail(gray < 0 || gray > 255);
		hist[gray]++;
		idxs[gray] = i;
		ptr += dim;
	}

	for (i = 0; i < n; i++) {
		hist[i] /= data_size;
	}

	memcpy(value, hist, n * sizeof(double));

	for (i = 0; i < n - 1; i++) {
		hist[i+1] += hist[i];
	}

	min = n + 1;
	front = 0;
	for (i = 0; i < n - 1; i++) {
		tmp = i * hist[i] + (n - i) * (1 - hist[i + 1]);
		if (tmp < min) {
			min = tmp;
			front = i;
			continue;
		}
		if (tmp == min) {
			if (hist[i] > hist[front]) {
				front = i;
			}
		}
	}
	tmp = -1;
	l = 0;
	for (i = 0; i < front; i++) {
		if (value[i] > tmp) {
			tmp = value[i];
			l = i;
		}
	}

	tmp = -1;
	r = n - 1;
	for (i = front; i < n; i++) {
		if (value[i] > tmp) {
			tmp = value[i];
			r = i;
		}
	}
	if (idxs[l] == -1)
		l = r;
	if (idxs[r] == -1)
		r = l;

	ptr = means;
	ptr1 = data + idxs[l] * dim;
	for (i = 0; i < dim; i++) {
		*ptr = *ptr1;
		ptr++;
		ptr1++;
	}
	ptr1 = data + idxs[r] * dim;
	for (i = 0; i < dim; i++) {
		*ptr = *ptr1;
		ptr++;
		ptr1++;
	}
}

static void
bg_model_adaptive_init_rgb_3_means(double *means, int dim, int k, const double *data, int data_size)
{
	const int n = 256;
	double hist[n], value[n];
	int idxs[n];
	int i, l, r, front, gray, i_begin, i_end;
	double min, tmp, *ptr;
	double const *ptr1;
	memset(hist, 0, sizeof(double) * n);

	ptr1 = data;

	for (i = 0; i < n; i++) {
		idxs[i] = -1;
	}

	for (i = 0; i < data_size; i++) {
		gray = (int)(0.299 * ptr1[0] + 0.587 * ptr1[1] + 0.114 * ptr1[2]);
		return_if_fail(gray < 0 || gray > 255);
		hist[gray]++;
		idxs[gray] = i;
		ptr += dim;
	}

	for (i = 0; i < n; i++) {
		hist[i] /= data_size;
	}

	memcpy(value, hist, n * sizeof(double));

	for (i = 0; i < n - 1; i++) {
		hist[i+1] += hist[i];
	}

	min = n + 1;
	front = 0;
	for (i = 0; i < n - 1; i++) {
		tmp = i * hist[i] + (n - i) * (1 - hist[i + 1]);
		if (tmp < min) {
			min = tmp;
			front = i;
			continue;
		}
		if (tmp == min) {
			if (hist[i] > hist[front]) {
				front = i;
			}
		}
	}

	tmp = -1;
	l = 0;
	for (i = 0; i < front; i++) {
		if (value[i] > tmp) {
			tmp = value[i];
			l = i;
		}
	}

	tmp = -1;
	r = n - 1;
	for (i = front; i < n; i++) {
		if (value[i] > tmp) {
			tmp = value[i];
			r = i;
		}
	}
	if (idxs[l] == -1)
		l = r;
	if (idxs[r] == -1)
		r = l;

	ptr = means;

	if (front > n / 2) {
		i_begin = 0;
		i_end   = front;
		if (idxs[r] == -1) return;
		ptr1 = data + idxs[r] * dim;
		for (i = 0; i < dim; i++) {
			*ptr = *ptr1;
			ptr++;
			ptr1++;
		}
	} else {
		i_begin = front;
		i_end   = n;
		if (idxs[l] == -1) return;
		ptr1 = data + idxs[l] * dim;
		for (i = 0; i < dim; i++) {
			*ptr = *ptr1;
			ptr++;
			ptr1++;
		}
	}
	min = n + 1;
	front = i_begin;
	for (i = i_begin; i < i_end - 1; i++) {
		tmp = i * hist[i] + (n - i) * (1 - hist[i + 1]);
		if (tmp < min) {
			min = tmp;
			front = i;
			continue;
		}
		if (tmp == min) {
			if (hist[i] > hist[front]) {
				front = i;
			}
		}
	}

	tmp = -1;
	l = i_begin;
	for (i = i_begin; i < front; i++) {
		if (value[i] > tmp) {
			tmp = value[i];
			l = i;
		}
	}

	tmp = -1;
	r = i_end - 1;
	for (i = front; i < i_end - 1; i++) {
		if (value[i] > tmp) {
			tmp = value[i];
			r = i;
		}
	}

	if (idxs[l] == -1)
		l = r;
	if (idxs[r] == -1)
		r = l;
	if (idxs[l] == -1 || idxs[r] == -1) return;
	ptr1 = data + idxs[l] * dim;
	for (i = 0; i < dim; i++) {
		*ptr = *ptr1;
		ptr++;
		ptr1++;
	}
	ptr1 = data + idxs[r] * dim;
	for (i = 0; i < dim; i++) {
		*ptr = *ptr1;
		ptr++;
		ptr1++;
	}
}

struct bg_model *
bg_model_new(enum bg_model_type model)
{
	struct bg_model *ptr;
	ptr = malloc(sizeof(struct bg_model));
	return_val_if_fail(ptr != NULL, ptr);
	ptr->model_type = model;
	ptr->buff = new_circle_buff();
	return_val_if_fail(ptr->buff != NULL, ptr);
	switch (model) {
	case BG_MODEL_MOG:
		ptr->bg_model_ctx = (void*)em_mog_new();
		em_mog_set_dimension(ptr->bg_model_ctx, 3);
		break;
	case BG_MODEL_K_MEANS:
		ptr->bg_model_ctx = (void*)k_means_new();
		k_means_set_dimension(ptr->bg_model_ctx, 3);
		break;
	}

	return ptr;
}

void
bg_model_init(struct bg_model *bg)
{
	return_if_fail(bg != NULL);
	switch (bg->colorspace) {
	case BG_COLORSPACE_RGB:
		bg->buff->dim = 3;
		break;
	}

	switch (bg->model_type) {
	case BG_MODEL_MOG:
		em_mog_init((struct em_mog_context*)bg->bg_model_ctx);
		break;
	case BG_MODEL_K_MEANS:
		k_means_init_((struct k_means_context*)bg->bg_model_ctx);
		break;
	default:
		break;
	}
	init_circle_buff(bg->buff);
}

void
bg_model_clear(struct bg_model *bg)
{
	return_if_fail(bg != NULL);
	free_circle_buff(&(bg->buff));

	switch (bg->model_type) {
	case BG_MODEL_MOG:
		em_mog_free((struct em_mog_context **)&(bg->bg_model_ctx));
		break;
	case BG_MODEL_K_MEANS:
		k_means_free((struct k_means_context**)&(bg->bg_model_ctx));
		break;
	default:
		break;
	}
}

void
bg_model_free(struct bg_model **bg)
{
	return_if_fail(bg != NULL);
	return_if_fail((*bg) != NULL);

	bg_model_clear(*bg);
	free(*bg);
	*bg = NULL;
}

/* Set colorspace. The only colorspace supported is BG_COLORSPACE_RGB. */
void
bg_model_set_colorspace(struct bg_model *bg, enum bg_colorspace colorspace)
{
	return_if_fail(bg != NULL);
	bg->colorspace = colorspace;

	switch (colorspace) {
	case BG_COLORSPACE_RGB:
		bg->buff->dim = 3;
		break;
	}
}

int
bg_model_get_colorspace(struct bg_model *bg)
{
	return_val_if_fail(bg != NULL, -1);
	return bg->colorspace;
}

/* Set underlying model parameters using param string. */
void
bg_model_set_parameters(struct bg_model *bg, const char *param)
{
	return_if_fail(bg != NULL);
	return_if_fail(bg->bg_model_ctx != NULL);

	int nth;
	double mean[3], cmx[9], w, epsilon;
	void *ptr = bg->bg_model_ctx;
	char *p1, *p2;
	int i, n_m, n_cmx;

	n_m   = 3;
	n_cmx = 9;
	switch (bg->model_type) {
		case BG_MODEL_MOG:
		{
			struct em_mog_context* ctx = (struct em_mog_context*)ptr;

			if ((p1 = strstr(param, bg_em_cmd[0])) != NULL) {
				p1 += strlen(bg_em_cmd[0]);
				errno = 0;
				nth = strtol(p1, &p2, 10);
				if (errno != 0)
					goto mog_err;

				em_mog_set_ngaussians(ctx, nth);
				goto quit;
			}
			if ((p1 = strstr(param, bg_em_cmd[1])) != NULL) {
				p1 += strlen(bg_em_cmd[1]);
				errno = 0;
				nth = strtol(p1, &p2, 10);
				if (errno != 0)
					goto mog_err;
				p1 = p2;
				for (i = 0; i < n_m; i++) {
					mean[i] = strtod(p1, &p2);
					if (errno == ERANGE)
						goto mog_err;
					p1 = p2;
				}
				em_mog_set_nth_mean(ctx, nth, mean);
				goto quit;
			}

			if ((p1 = strstr(param, bg_em_cmd[2])) != NULL) {
				p1 += strlen(bg_em_cmd[2]);
				errno = 0;
				nth = strtol(p1, &p2, 10);
				if (errno != 0)
					goto mog_err;
				p1 = p2;

				w = strtod(p1, &p2);
				if (errno == ERANGE)
					goto mog_err;

				em_mog_set_nth_weight(ctx, nth, w);
				goto quit;
			}
			if ((p1 = strstr(param, bg_em_cmd[3])) != NULL) {
				p1 += strlen(bg_em_cmd[3]);
				errno = 0;
				nth = strtol(p1, &p2, 10);
				if (errno != 0)
					goto mog_err;
				p1 = p2;
				for (i = 0; i < n_cmx; i++) {
					cmx[i] = strtod(p1, &p2);
					if (errno == ERANGE)
						goto mog_err;
					p1 = p2;
				}
				em_mog_set_nth_covariance_matrix(ctx, nth, cmx);
				goto quit;
			}
			if ((p1 = strstr(param, bg_em_cmd[4])) != NULL) {
				p1 += strlen(bg_em_cmd[4]);
				errno = 0;
				epsilon = strtod(p1, &p2);
				if (errno == ERANGE)
					goto mog_err;
				em_mog_set_error(ctx, epsilon);
				goto quit;
			}
			if ((p1 = strstr(param, bg_em_cmd[5])) != NULL) {
				p1 += strlen(bg_em_cmd[5]);
				errno = 0;
				double noice_level = strtod(p1, &p2);
				if (errno == ERANGE)
					goto mog_err;
				em_mog_set_noise_level(ctx, noice_level);
				goto quit;
			}
mog_err:
			fprintf(stderr, "Error in command : %s for mixture of Gaussian model\n", param);
		}
			break;
		case BG_MODEL_K_MEANS: {
			struct k_means_context *ctx = (struct k_means_context*)ptr;

			if ((p1 = strstr(param, bg_k_means_cmd[0])) != NULL) {
				p1 += strlen(bg_k_means_cmd[0]);
				errno = 0;
				nth = strtol(p1, &p2, 10);
				if (errno == ERANGE)
					goto k_means_err;
				k_means_set_nmean(ctx, nth);
				goto quit;
			}
			if ((p1 = strstr(param, bg_k_means_cmd[1])) != NULL) {
				p1 += strlen(bg_k_means_cmd[1]);
				errno = 0;
				nth = strtol(p1, &p2, 10);
				if (errno == ERANGE)
					goto k_means_err;
				p1 = p2;
				for (i = 0; i < n_m; i++) {
					mean[i] = strtod(p1, &p2);
					if (errno == ERANGE)
						goto k_means_err;
					p1 = p2;
				}
				k_means_set_nth_mean(ctx, nth, mean);
				goto quit;
			}
k_means_err:
			fprintf(stderr, "Error in command : %s for k-means model\n", param);
		}
			break;
		default:
			fprintf(stderr, "Error in command: %s\n", param);
			break;
	}
quit:;
}

/* Set depth of the pixel value history cyclic buffer. */
void bg_model_set_history_size(struct bg_model *bg, int size)
{
	return_if_fail(bg != NULL);
	return_if_fail(bg->buff != NULL);
	bg->buff->buff_size = size;
}

int
bg_model_get_history_size(struct bg_model *bg)
{
	return_val_if_fail(bg != NULL, -1);
	return_val_if_fail(bg->buff != NULL, -1);
	return bg->buff->buff_size;
}

/* Update the history with a new pixel value. Arguments [x,y,z] stay for
 * [R,G,B] values in BG_COLORSPACE_RGB.
 */
void
bg_model_history_add(struct bg_model *bg, int x, int y, int z)
{
	double data[3];
	return_if_fail(bg != NULL);
	return_if_fail(bg->buff != NULL);
	data[0] = x;
	data[1] = y;
	data[2] = z;
	write_circle_buff(bg->buff, data);
}

/* Reset the history and forget all previous pixel values. */
void
bg_model_history_reset(struct bg_model *bg)
{
	return_if_fail(bg != NULL);
	return_if_fail(bg->buff != NULL);
	reset_circle_buff(bg->buff);
}

/* Acquire background model for the single pixel using it's collected historical values. */
void
bg_model_acquire(struct bg_model *bg)
{
	return_if_fail(bg != NULL);
	return_if_fail(bg->buff != NULL);
	return_if_fail(bg->bg_model_ctx != NULL);

	void *ptr = bg->bg_model_ctx;
	struct circle_buff *tmp = bg->buff;

	switch (bg->model_type) {
	case BG_MODEL_MOG: {
		struct em_mog_context *ctx = (struct em_mog_context*)ptr;
		em_mog_converge(ctx, tmp->buff, tmp->buff_size,
				ctx->epsilon);
	}
		break;
	case BG_MODEL_K_MEANS: {
		struct k_means_context *ctx = (struct k_means_context *)ptr;
		k_means_converge(ctx, tmp->buff, tmp->buff_size, 0.0);
	}
		break;
	}
}

/* Update background model using a new pixel value. Arguments [x,y,z]
 * stay for [R,G,B] values in BG_COLORSPACE_RGB. */
void
bg_model_update(struct bg_model *bg, int x, int y, int z, int n_model)
{
	return_if_fail(bg != NULL);
	return_if_fail(bg->bg_model_ctx != NULL);
	double data[3];
	data[0] = x;
	data[1] = y;
	data[2] = z;
	write_circle_buff(bg->buff, data);
	switch (bg->model_type) {
		case BG_MODEL_MOG: {
			struct em_mog_context *ptr = (struct em_mog_context *)bg->bg_model_ctx;
			em_mog_update_nth_model(ptr, data, n_model);
		}
			break;
		case BG_MODEL_K_MEANS: {
			struct k_means_context *ptr = (struct k_means_context *)bg->bg_model_ctx;
			k_means_update(ptr, data);
		}
			break;
		default:
			break;
	}
}

/* Reset the model to initial state. This resets underlaying mathematical model
 * to initial state too. */
void
bg_model_reset(struct bg_model *bg)
{
	return_if_fail(bg != NULL);
	return_if_fail(bg->buff != NULL);
	return_if_fail(bg->bg_model_ctx != NULL);

	void *ptr = bg->bg_model_ctx;
	double *data = bg->buff->buff;
	int data_size, dim;
	data_size = bg->buff->tail;
	dim = 3;

	switch (bg->model_type) {
	case BG_MODEL_MOG: {
		struct em_mog_context* ctx = (struct em_mog_context*)ptr;
		em_mog_reset(ctx, data, data_size);
		if (ctx->ngauss >= 2)
			bg_model_adaptive_init_rgb_2_means(ctx->means, ctx->dim, 2, data, data_size);
	}
		break;
	case BG_MODEL_K_MEANS: {
		struct k_means_context* ctx = (struct k_means_context*)ptr;
		k_means_reset(ctx, data, data_size);
	}
		break;
	}
}

/* Determine, according to current background model, if a pixel [x,y,z]
 * represents background or not. Arguments [x,y,z] stay for [R,G,B] values in
 * BG_COLORSPACE_RGB */
int
bg_model_pixel_to_match_model(struct bg_model *bg, int x, int y, int z)
{
	return_val_if_fail(bg != NULL, BG_ERROR);
	return_val_if_fail(bg->buff != NULL, BG_ERROR);
	return_val_if_fail(bg->bg_model_ctx != NULL, BG_ERROR);
	double data[3] = {x, y, z};

	switch (bg->model_type) {
		case BG_MODEL_MOG: {
			struct em_mog_context *ctx = (struct em_mog_context *)bg->bg_model_ctx;
			return em_mog_sample_to_match_model(ctx, data);
		}
			break;
		case BG_MODEL_K_MEANS: {
			struct k_means_context *ctx = (struct k_means_context *)bg->bg_model_ctx;
			return k_means_sample_to_match_model(ctx, data);
		}
			break;
		default:
			return BG_ERROR;
			break;
	}
}
